import os
import traceback

import hypertrace.agent.filter
from hypertrace.agent.filter.registry import TYPE_HTTP, TYPE_RPC
from opentelemetry.trace import Span
from traceableai.config.config import Config
from hypertrace.agent.custom_logger import get_custom_logger # pylint:disable=C0413,C0411,C0412

logger = get_custom_logger(__name__)

_LIBTRACEABLE_AVAILABLE = False
_NOT_BODY = False
_IS_BODY = True

try:
    import traceableai.filter._libtraceable as _libtraceable  # pylint: disable=E0611,E0401,bad-option-value
    _LIBTRACEABLE_AVAILABLE = True
except Exception as error:  # pylint: disable=broad-except
    _LIBTRACEABLE_AVAILABLE = False
    traceback.print_tb(error.__traceback__)
    logger.warning(error)
    logger.warning(
        "error loading traceable extension - libtraceable filter disabled")


_HEADER_PREFIXES = {
    TYPE_HTTP: 'http.request.header',
    TYPE_RPC: 'rpc.request.metadata'
}

_BODY_PREFIXES = {
    TYPE_HTTP: 'http.request.body',
    TYPE_RPC: 'rpc.request.body'
}


class Traceable(hypertrace.agent.filter.Filter):

    def __init__(self):
        if _LIBTRACEABLE_AVAILABLE:
            self.libtraceable_config = _create_libtraceable_config()
            p_libtraceable = _libtraceable.ffi.new("traceable_libtraceable*")
            result_code = _libtraceable.lib.traceable_new_libtraceable(
                self.libtraceable_config, p_libtraceable)
            self.libtraceable_available = (
                result_code == _libtraceable.lib.TRACEABLE_SUCCESS)
            if self.libtraceable_available:
                logger.debug("libtraceable available")
                self.libtraceable = p_libtraceable[0]
                result = _libtraceable.lib.traceable_start_libtraceable(
                    self.libtraceable)
                if result != _libtraceable.lib.TRACEABLE_SUCCESS:
                    logger.debug("Failed to start libtraceable")
                    self.libtraceable_available = False
        else:
            self.libtraceable_available = False

    def evaluate_url_and_headers(self, span: Span, url: str, headers: dict, request_type) -> bool:
        if not self.libtraceable_available:
            return False

        pairs = _add_header_and_span_attributes(span, headers, request_type)

        # add url as part of header processing
        pairs.append((_libtraceable.ffi.new("char[]", b"http.url"),
                      _libtraceable.ffi.new("char[]", f"{url}".encode('ascii'))))

        num_headers = len(pairs)
        attributes = _libtraceable.ffi.new("traceable_attributes*")
        attribute_array = _libtraceable.ffi.new(
            "traceable_attribute[]", num_headers)
        attributes.count = num_headers

        for index, pair in enumerate(pairs, start=0):
            attribute_array[index].key = pair[0]
            attribute_array[index].value = pair[1]
        attributes.attribute_array = attribute_array

        return self._evaluate_attributes(span, attributes, _NOT_BODY)

    def evaluate_body(self, span: Span, body, headers: dict, request_type) -> bool:
        if not self.libtraceable_available:
            return False

        if not body or len(str(body)) == 0:
            return False

        pairs = _add_header_and_span_attributes(span, headers, request_type)
        prefix = _BODY_PREFIXES[request_type]
        try:
            if isinstance(body, bytes):
                pairs.append((_libtraceable.ffi.new("char[]", prefix.encode('ascii')),
                              _libtraceable.ffi.new("char[]", body)))
            else:
                pairs.append((_libtraceable.ffi.new("char[]", prefix.encode('ascii')),
                              _libtraceable.ffi.new("char[]", f"{body}".encode('ascii'))))
        except Exception as exception: # pylint:disable=W0703
            logger.error("Error adding body data to extension call: %s", exception)

        num_attrs = len(pairs)
        attributes = _libtraceable.ffi.new("traceable_attributes*")
        attribute_array = _libtraceable.ffi.new(
            "traceable_attribute[]", num_attrs)
        attributes.count = num_attrs

        for index, pair in enumerate(pairs, start=0):
            attribute_array[index].key = pair[0]
            attribute_array[index].value = pair[1]
        attributes.attribute_array = attribute_array

        return self._evaluate_attributes(span, attributes, _IS_BODY)

    def _evaluate_attributes(self, span, attributes, is_body) -> bool:
        result = _libtraceable.ffi.new("traceable_process_request_result*")
        result_code = _libtraceable.lib.TRACEABLE_SUCCESS
        if is_body:
            result_code = _libtraceable.lib.traceable_process_request_body(
                self.libtraceable, attributes[0], result)  # pylint:disable=C0301
        else:
            result_code = _libtraceable.lib.traceable_process_request_headers(
                self.libtraceable, attributes[0], result)

        if result_code != _libtraceable.lib.TRACEABLE_SUCCESS:
            logger.debug("traceable_process_request fail!")
            return False

        _add_span_attributes(span, result)

        _result_code = _libtraceable.lib.traceable_delete_process_request_result_data(
            result[0])
        if result.block:
            return True
        return False


def _add_span_attributes(span: Span, process_request_result):
    attr_count = process_request_result.attributes.count
    for i in range(attr_count):
        key_attr = process_request_result.attributes.attribute_array[i].key
        value_attr = process_request_result.attributes.attribute_array[i].value
        if key_attr == _libtraceable.ffi.NULL or value_attr == _libtraceable.ffi.NULL:
            continue

        key = _libtraceable.ffi.string(key_attr)
        value = _libtraceable.ffi.string(value_attr)

        span.set_attribute(key.decode('utf-8', 'backslashescape'),
                           value.decode('utf-8', 'backslashescape'))


def _add_header_and_span_attributes(span: Span, headers: dict, request_type) -> []:
    pairs = []
    ip_address = span._readable_span().attributes.get('net.peer.ip')  # pylint:disable=W0212
    if ip_address:
        pairs.append((_libtraceable.ffi.new("char[]", b"net.peer.ip"),
                      _libtraceable.ffi.new("char[]", f"{ip_address}".encode('ascii'))))

    header_prefix = _HEADER_PREFIXES[request_type]
    for header_key, header_value in headers.items():
        pairs.append((_libtraceable.ffi.new("char[]", f"{header_prefix}.{header_key}".lower().encode('ascii')),
                      _libtraceable.ffi.new("char[]", f"{header_value}".encode('ascii'))))

    return pairs


def _create_libtraceable_config():
    traceable_config = Config().config
    should_debug = 0
    if os.environ.get("TA_LOG_LEVEL", "").upper() == "DEBUG":
        should_debug = 1
    if traceable_config.blocking_config.debug_log.value is True:
        should_debug = 1

    libtraceable_config = _libtraceable.lib.init_libtraceable_config()

    libtraceable_config.log_config.mode = should_debug

    # If blocking is disabled we will never attempt to initialize the traceable filter
    # so we can safely set to 1 here
    libtraceable_config.blocking_config.enabled = 1

    traceable_opa = traceable_config.opa
    opa_server_url = _libtraceable.ffi.new(
        "char[]", f"{traceable_opa.endpoint.value}".encode('ascii'))

    opa_enabled = (0 if not traceable_opa.enabled.value is True else 1)
    libtraceable_config.blocking_config.opa_config.enabled = opa_enabled
    libtraceable_config.blocking_config.opa_config.opa_server_url = opa_server_url
    libtraceable_config.blocking_config.opa_config.log_to_console = 1
    libtraceable_config.blocking_config.opa_config.debug_log = should_debug
    libtraceable_config.blocking_config.opa_config.min_delay = traceable_opa.poll_period_seconds.value
    libtraceable_config.blocking_config.opa_config.max_delay = traceable_opa.poll_period_seconds.value

    modsecurity_enabled = (
        1 if traceable_config.blocking_config.modsecurity.enabled.value is True else 0)

    libtraceable_config.blocking_config.modsecurity_config.enabled = modsecurity_enabled

    rb_enabled = 1 if traceable_config.blocking_config.region_blocking.enabled.value is True else 0
    libtraceable_config.blocking_config.rb_config.enabled = rb_enabled

    libtraceable_config.blocking_config.evaluate_body = (
        1 if traceable_config.blocking_config.evaluate_body.value is True else 0)
    libtraceable_config.blocking_config.skip_internal_request = (
        1 if traceable_config.blocking_config.skip_internal_request.value is True else 0)  # pylint:disable=C0301

    remote_config_enabled = (
        1 if traceable_config.blocking_config.remote_config.enabled.value is True else 0)
    libtraceable_config.remote_config.enabled = remote_config_enabled

    cert_file = _libtraceable.ffi.new("char[]", "".encode('ascii'))
    libtraceable_config.remote_config.cert_file = cert_file

    remote_config_endpoint = (
        traceable_config.blocking_config.remote_config.endpoint.value)
    remote_url = _libtraceable.ffi.new(
        "char[]", f"{remote_config_endpoint}".encode('ascii'))
    libtraceable_config.remote_config.remote_endpoint = remote_url

    remote_poll_period_seconds = (
        traceable_config.blocking_config.remote_config.poll_period_seconds.value)
    libtraceable_config.remote_config.poll_period_sec = remote_poll_period_seconds

    service_name = _libtraceable.ffi.new( "char[]", "".encode('ascii'))
    libtraceable_config.agent_config.service_name = service_name
    return libtraceable_config



import os
import traceback
from logging import getLogger

import cffi

logger = getLogger(__name__)

def build(libtraceable_dir):
    try:
        logger.debug('attempting to build libtraceable ffi')
        ffi = cffi.FFI()
        include_path = os.path.join(libtraceable_dir, 'libtraceable.h')
        logger.debug(
            "Generating libtraceable at include path: %s", include_path)
        static_lib_path = os.path.join(libtraceable_dir, "libtraceable_full.a")
        ffi.set_source("traceableai.filter._libtraceable",
                       # Since we are calling a fully built library directly no custom source
                       # is necessary. We need to include the .h files, though, because behind
                       # the scenes cffi generates a .c file which contains a Python-friendly
                       # wrapper around each of the functions.

                       # We need to use the absolute path for libtraceable.h to avoid path differences
                       # when _libtraceable is referenced from __pycache__ directory
                       '#include "{}"'.format(include_path),
                       # The important thing is to include the pre-built lib in the list of
                       # libraries we are linking against:

                       # static lib uses "extra_objects" keyword
                       extra_objects=[static_lib_path],
                       # needed when using the static lib
                       libraries=["stdc++"],
                       )

        preprocessed_header_path = os.path.join(
            os.path.dirname(__file__), "libtraceable.h.preprocessed")
        with open(preprocessed_header_path) as preprocessed_header_file:
            ffi.cdef(preprocessed_header_file.read())

        src_dir = os.path.join(os.path.dirname(__file__), '..', '..')
        logger.debug(src_dir)
        ffi.compile(verbose=True, tmpdir=src_dir)
        logger.debug('successfully built libtraceable ffi')
        return True
    except Exception:  # pylint: disable=broad-except
        logger.error('failed to build libtraceable ffi, %s',
                     traceback.format_exc())
    return False



import os
import traceback
from contextlib import contextmanager

import distro

# These imports have to be first to respect TA_ env vars in hypertrace agent
from hypertrace import env_var_settings

from traceableai.config.config import Config
from traceableai.config_wrapper import ConfigWrapper
from traceableai.hypertrace_env_override import add_deprecation_to_ht_env

env_var_settings.ENV_VAR_PREFIXES.insert(0, 'TA')
add_deprecation_to_ht_env()

from hypertrace import agent  # pylint:disable=C0413,C0411,C0412
from hypertrace.agent.filter.registry import Registry # pylint:disable=C0413,C0411,C0412
from hypertrace.agent.custom_logger import get_custom_logger # pylint:disable=C0413,C0411,C0412
from hypertrace.agent import constants # pylint:disable=C0413,C0411,C0412

from traceableai.version import __version__ # pylint:disable=C0413

logger = get_custom_logger(__name__)


# We need to override the version that is reported to the platform from hypertrace
constants.TELEMETRY_SDK_VERSION = __version__

class Agent(agent.Agent):
    def __init__(self):
        super().__init__()
        self.is_lambda = False
        logger.debug("Platform: %s", distro.id())
        logger.debug("Platform version: %s", distro.version())
        logger.debug('TraceableAI Agent version: %s', __version__)
        logger.debug("successfully initialized traceableai agent")

        if hasattr(os, 'register_at_fork'):
            logger.info('Registering after_in_child handler.')
            os.register_at_fork(after_in_child=self.post_fork)  # pylint:disable=E1101

    def post_fork(self):
        logger.info("In post fork hook")
        logger.info("Calling add traceable filter during post fork")
        self._init.post_fork()
        self.add_traceable_filter()

    @contextmanager
    def edit_config(self):
        """Used by end users to modify the config"""
        with super().edit_config() as ht_config:
            bundled_config = ConfigWrapper(ht_config, Config().config)
            config_dict = bundled_config.current_config()
            yield config_dict
            bundled_config.apply_modifications(config_dict)


    def add_traceable_filter(self):  #pylint:disable=R0201
        logger.debug("in add_traceable_filter")
        if self.is_lambda:
            logger.info('Not loading blocking extension - currently unsupported in lambda')
            return
        # We need to do a local import so that the extension is not loaded in a parent process
        from traceableai.filter.traceable import Traceable, _LIBTRACEABLE_AVAILABLE  # pylint:disable=C0413,C0412,C0415
        if not _LIBTRACEABLE_AVAILABLE:
            logger.info("libtraceable unavailable, skipping filter registration")
            return
        if Config().config.blocking_config.enabled.value is not True:
            logger.info("Not adding libtraceable filter - blocking is not enabled")
            return
        try:
            Registry().register(Traceable)
            logger.debug("successfully initialized traceable filter")
        except Exception as exc: # pylint:disable=W0703
            logger.debug(''.join(traceback.format_exception(None, exc, exc.__traceback__)))
            logger.info("failed to register traceable filter")



"""Exposes wrapper for user to re-init agent in prefork web servers"""
def post_fork(server, worker):
    """Used to reinitialize exporter & processors in separate worker processes"""
    from traceableai.agent import Agent  # pylint:disable=C0411,C0415
    server.log.info("Add post hook %s", worker.pid)
    Agent()._init.post_fork()  # pylint:disable=W0212
    Agent().add_traceable_filter()



from traceableai.config import traceable_hypertrace_config


# To simplify usage for end users
# we bundle the traceable & hypertrace configs into a single dict
# Then once modifications are complete, apply the dict values back onto the original protobuf
# the primary benefit is users no longer have to use protobuf fields/values and can modify the config with primitives
# Additionally we don't have to expose the 2 configs independently, instead they are yielded as a single config
class ConfigWrapper:  # pylint:disable=R0903
    def __init__(self, ht_config, ta_config):
        self.traceable = ta_config
        self.hypertrace = ht_config

    def apply_modifications(self, modified_config_dict):
        # blocking config
        self.traceable.blocking_config.enabled.value = modified_config_dict["blocking_config"]["enabled"]
        self.traceable.blocking_config.debug_log.value = modified_config_dict["blocking_config"]["debug_log"]
        self.traceable.blocking_config.evaluate_body.value = modified_config_dict["blocking_config"]["evaluate_body"]
        self.traceable.blocking_config.modsecurity.enabled.value = \
            modified_config_dict["blocking_config"]["modsecurity"]["enabled"]
        self.traceable.blocking_config.skip_internal_request.value = \
            modified_config_dict["blocking_config"]["skip_internal_request"]
        self.traceable.blocking_config.region_blocking.enabled.value = \
            modified_config_dict["blocking_config"]["region_blocking"]["enabled"]
        self.traceable.blocking_config.remote_config.enabled.value = \
            modified_config_dict["blocking_config"]["remote_config"]["enabled"]
        self.traceable.blocking_config.remote_config.endpoint.value =\
            modified_config_dict["blocking_config"]["remote_config"]["endpoint"]
        self.traceable.blocking_config.remote_config.poll_period_seconds.value = \
            modified_config_dict["blocking_config"]["remote_config"]["poll_period_seconds"]
        # opa
        self.traceable.opa.enabled.value = modified_config_dict["opa"]["enabled"]
        self.traceable.opa.endpoint.value = modified_config_dict["opa"]["endpoint"]
        self.traceable.opa.poll_period_seconds.value = modified_config_dict["opa"]["poll_period_seconds"]

        # hypertrace top level
        self.hypertrace.enabled = modified_config_dict["enabled"]
        self.hypertrace.service_name = modified_config_dict["service_name"]

        proto_propagation_formats = []
        for prop_format in modified_config_dict["propagation_formats"]:
            proto_propagation_formats.append(traceable_hypertrace_config.PropagationFormat.Value(prop_format))
        self.hypertrace.propagation_formats = proto_propagation_formats

        # reporting
        reporting = modified_config_dict["reporting"]
        self.hypertrace.reporting.endpoint = reporting["endpoint"]
        self.hypertrace.reporting.secure = reporting["secure"]
        self.hypertrace.reporting.trace_reporter_type = \
            traceable_hypertrace_config.TraceReporterType.Value(reporting["trace_reporter_type"])

        # data capture
        data_capture = modified_config_dict["data_capture"]
        self.hypertrace.data_capture.http_headers.request.value = data_capture["http_headers"]["request"]
        self.hypertrace.data_capture.http_headers.response.value = data_capture["http_headers"]["response"]
        self.hypertrace.data_capture.http_body.request.value = data_capture["http_body"]["request"]
        self.hypertrace.data_capture.http_body.response.value = data_capture["http_body"]["response"]
        self.hypertrace.data_capture.rpc_metadata.request.value = data_capture["rpc_metadata"]["request"]
        self.hypertrace.data_capture.rpc_metadata.response.value = data_capture["rpc_metadata"]["response"]
        self.hypertrace.data_capture.rpc_body.request.value = data_capture["rpc_body"]["request"]
        self.hypertrace.data_capture.rpc_body.response.value = data_capture["rpc_body"]["response"]
        self.hypertrace.data_capture.body_max_size_bytes = data_capture["body_max_size_bytes"]

    def current_config(self):
        # We should revisit the Hypertrace config loader, current it prevents us from doing:
        # MessageToDict(ht_config)
        # Then this could become:
        # current_config = {}.merge(MessageToDict(ht_config).merge(MessageToDict(ta_config))
        # Similar in apply modifications we could then just use AgentConfig.CopyFrom
        propagation_formats = []
        for prop_format in self.hypertrace.propagation_formats:
            propagation_formats.append(traceable_hypertrace_config.PropagationFormat.Name(prop_format))

        return {
            'enabled': self.hypertrace.enabled,
            'propagation_formats': propagation_formats,
            'service_name': self.hypertrace.service_name,
            'reporting': {
                'endpoint': self.hypertrace.reporting.endpoint,
                'secure': self.hypertrace.reporting.secure,
                'trace_reporter_type': traceable_hypertrace_config.TraceReporterType.Name(self.hypertrace.reporting.trace_reporter_type), # pylint:disable=C0301
            },
            'data_capture': {
                'http_headers': {
                    'request': self.hypertrace.data_capture.http_headers.request.value,
                    'response': self.hypertrace.data_capture.http_headers.response.value,
                },
                'http_body': {
                    'request': self.hypertrace.data_capture.http_body.request.value,
                    'response': self.hypertrace.data_capture.http_body.response.value,
                },
                'rpc_metadata': {
                    'request': self.hypertrace.data_capture.rpc_metadata.request.value,
                    'response': self.hypertrace.data_capture.rpc_metadata.response.value,
                },
                'rpc_body': {
                    'request': self.hypertrace.data_capture.rpc_body.request.value,
                    'response': self.hypertrace.data_capture.rpc_metadata.response.value,
                },
                'body_max_size_bytes': self.hypertrace.data_capture.body_max_size_bytes,
            },
            'opa': {
                'enabled': self.traceable.opa.enabled.value,
                'endpoint': self.traceable.opa.endpoint.value,
                'poll_period_seconds': self.traceable.opa.poll_period_seconds.value
            },
            'blocking_config': {
                'enabled': self.traceable.blocking_config.enabled.value,
                'debug_log': self.traceable.blocking_config.debug_log.value,
                'evaluate_body': self.traceable.blocking_config.evaluate_body.value,
                'modsecurity': {
                    'enabled': self.traceable.blocking_config.modsecurity.enabled.value,
                },
                'skip_internal_request': self.traceable.blocking_config.skip_internal_request.value,
                'region_blocking': {
                    'enabled': self.traceable.blocking_config.region_blocking.enabled.value,
                },
                'remote_config': {
                    'enabled': self.traceable.blocking_config.remote_config.enabled.value,
                    'endpoint': self.traceable.blocking_config.remote_config.endpoint.value,
                    'poll_period_seconds': self.traceable.blocking_config.remote_config.poll_period_seconds.value,
                }
            }
        }



import os

from hypertrace import env_var_settings
from hypertrace.agent.custom_logger import get_custom_logger # pylint:disable=C0413,C0411,C0412
logger = get_custom_logger(__name__)

def ta_get_env_value(target_key):
    for prefix in env_var_settings.ENV_VAR_PREFIXES:
        env_var_key = f"{prefix}_{target_key}"
        if env_var_key in os.environ:
            if env_var_key.startswith('HT_'):
                logger.warning("[Deprecated] - HT_ prefixed environment "
                               "variables should be updated to instead be prefixed with TA_")
                logger.warning("[Deprecated] - Update from %s to TA_%s", env_var_key, target_key)
            return os.environ[env_var_key]
    return None


def add_deprecation_to_ht_env():
    env_var_settings.get_env_value = ta_get_env_value
